# pkpass_field_row.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from gi.repository import GLib, Gtk


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/pass_field_row.ui')
class PassFieldRow(Gtk.ListBoxRow):

    __gtype_name__ = 'PassFieldRow'

    label = Gtk.Template.Child()
    value = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self.value.set_use_markup(True)

    def set_label(self, label):
        if label and label.strip():
            self.label.set_text(label)
            self.label.show()
        else:
            self.label.hide()

    def set_value(self, value):
        value = str(value)
        value = re.sub(r'\<br\/>', '\n', value)
        value_has_links = re.search('</a>', value)

        if value_has_links:
            value = re.sub('&', '&amp;', value)

        else:
            value = GLib.markup_escape_text(value)

            # Create a link for URLs
            value = re.sub('(?:(https?://)|(www))(\S+)',
                           '<a href="https://\\2\\3">\\1\\2\\3</a>',
                           value)

            # Create a link for telephone numbers
            value = re.sub('(\+\d+[\(\)\-\d\s\.]+\d)',
                           '<a href="tel:\\1">\\1</a>',
                           value)

            # Create a link for e-mails
            value = re.sub('(\S+\@[\w\-]+\.\w+)',
                           '<a href="mailto:\\1">\\1</a>',
                           value)

        self.value.set_label(value)



# settings.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, tempfile

from gi.repository import Gio

from .digital_pass_list_store import SortingCriteria


class Settings:
    """
    """

    def __init__(self, application_id):
        self.__gsettings = Gio.Settings.new(application_id)

    def get_sorting_criteria(self):
        value = self.__gsettings\
            .get_value('default-sorting-criteria')\
            .get_string()

        return SortingCriteria.from_string(value)

    def set_sorting_criteria(self, sorting_criteria):
        self.__gsettings.set_string('default-sorting-criteria', sorting_criteria)




# espass.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .digital_pass import Barcode, Color, Date, DigitalPass, Image, \
                          PassDataExtractor, PassField, Date, TimeInterval


class EsPass():
    """
    A representation of an esPass pass
    """

    types = ['BOARDING',
              'COUPON',
              'EVENT',
              'LOYALTY',
              'VOUCHER']

    def __init__(self, pass_data, pass_images):
        self.__data = PassDataExtractor(pass_data)
        self.__images = pass_images

        self.__type = self.__data.get('type')

        self.__front_fields = []
        self.__hidden_fields = []
        fields = self.__data\
            .get_list('fields', EsPassField)

        for field in fields:
            if field.is_hidden():
                self.__hidden_fields.append(field)
            else:
                self.__front_fields.append(field)

        self.__validity_time_intervals = []
        timespan_dicts = self.__data\
            .get_list('validTimespans')

        for dict in timespan_dicts:
            timespan = TimeInterval.from_iso_strings(dict['from'], dict['to'])
            self.__validity_time_intervals.append(timespan)


    # Container

    def icon(self):
        return Image(self.__images['icon.png'])


    # Mandatory fields

    def type(self):
        return self.__type

    def description(self):
        return self.__data.get('description')

    def id(self):
        return self.__data.get('id')


    # Time info

    def valid_timespans(self):
        return self.__validity_time_intervals


    # Metadata

    def creator(self):
        return self.__data.get('creator') or _('Unknown')


    # Fields

    def front_fields(self):
        return self.__front_fields

    def hidden_fields(self):
        return self.__hidden_fields


    # Color

    def accent_color(self):
        return self.__data.get('accentColor', Color.from_css)


    # Barcode

    def barcode(self):
        return self.__data.get('barCode', Barcode)


class EsPassAdapter(DigitalPass):
    def __init__(self, pkpass):
        super().__init__()
        self.__adaptee = pkpass

    def adaptee(self):
        return self.__adaptee

    def additional_information(self):
        return self.__adaptee.hidden_fields()

    def background_color(self):
        return self.__adaptee.accent_color()

    def barcodes(self):
        return [self.__adaptee.barcode()]

    def creator(self):
        return self.__adaptee.creator()

    def description(self):
        return self.__adaptee.description()

    def expiration_date(self):
        now = Date.now()
        latest_expiration_date = None

        for interval in self.__adaptee.valid_timespans():
            latest_expiration_date = interval.end_time()
            if now in interval:
                break

        return latest_expiration_date

    def file_extension():
        return '.espass'

    def format(self):
        return 'espass'

    def icon(self):
        return self.__adaptee.icon()

    def is_updatable(self):
        return False

    def mime_type():
        return 'application/vnd.espass-espass+zip'

    def relevant_date(self):
        return None

    def unique_identifier(self):
        return '.'.join([self.__adaptee.id(), self.format()])

    def voided(self):
        return False


class EsPassField(PassField):
    """
    An EsPass Field
    """

    __slots__ = ('__hide')

    def __init__(self, espass_field_dictionary):

        super().__init__()

        self.__hide = False
        if 'hide' in espass_field_dictionary.keys():
            self.__hide = espass_field_dictionary['hide']

        if 'label' in espass_field_dictionary.keys():
            self._label = espass_field_dictionary['label']

        self._value = espass_field_dictionary['value']

    def is_hidden(self):
        return self.__hide



# pkpass_row.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GLib, Gtk

from .digital_pass_list_store import SortingCriteria
from .pass_icon import PassIcon
from .pass_row_header import PassRowHeader


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/pass_row.ui')
class PassRow(Gtk.ListBoxRow):

    __gtype_name__ = 'PassRow'

    icon = Gtk.Template.Child()
    title = Gtk.Template.Child()
    subtitle = Gtk.Template.Child()

    def __init__(self, a_pass):
        super().__init__()
        self.__pass = a_pass

        self.__header_text = ''
        self.__sorting_criteria = None

        if a_pass.icon():
            self.icon.set_image(a_pass.icon())

        if a_pass.background_color():
            self.icon.set_background_color(a_pass.background_color())


        description = GLib.markup_escape_text(a_pass.description())

        if self.__pass.has_expired():
            description = '<s>%s</s>' % description

        self.title.set_label(description)
        self.subtitle.set_label(a_pass.creator())

    def activate(self):
        self.emit('activate')

    def data(self):
        return self.__pass

    def header_text(self):
        return self.__header_text

    def hide_header(self):
        self.set_header(None)

    def show_header(self):
        if not self.__header_text:
            return

        header = PassRowHeader(self.__header_text)
        self.set_header(header)

    def style(self):
        return self.__pass.style()

    def update_header_text_for(self, sorting_criteria):
        if self.__sorting_criteria == sorting_criteria:
            return

        self.__header_type = sorting_criteria

        if sorting_criteria == SortingCriteria.CREATOR:
            self.__header_text = self.__pass.creator()

        elif sorting_criteria == SortingCriteria.DESCRIPTION:
            self.__header_text = self.__pass.description()[0].upper()

        elif sorting_criteria == SortingCriteria.EXPIRATION_DATE:
            row_date = self.__pass.expiration_date()
            self.__header_text = row_date.as_relative_pretty_string() \
                                 if row_date else _('Without expiration date')

        elif sorting_criteria == SortingCriteria.RELEVANT_DATE:
            row_date = self.__pass.relevant_date()
            self.__header_text = row_date.as_relative_pretty_string() \
                                 if row_date else _('Without relevant date')

        else:
            self.__header_text = ''



# pass_icon.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, Graphene, Gsk, Gtk

from .digital_pass import Color


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/pass_icon.ui')
class PassIcon(Gtk.Box):

    __gtype_name__ = 'PassIcon'

    BORDER_RADIUS = 6
    BORDER_WIDTH = 0
    ICON_SIZE = 45

    def __init__(self):
        super().__init__()
        self.__background_color = None
        self.__image = None

        self.props.height_request = PassIcon.ICON_SIZE + PassIcon.BORDER_WIDTH
        self.props.width_request = PassIcon.ICON_SIZE + PassIcon.BORDER_WIDTH

    def __draw_background(self, snapshot):
        if not self.__background_color:
            return

        rectangle = Graphene.Rect()
        rectangle.init(0, 0, PassIcon.ICON_SIZE + PassIcon.BORDER_WIDTH,
                             PassIcon.ICON_SIZE + PassIcon.BORDER_WIDTH)

        rounded_rectangle = Gsk.RoundedRect()
        rounded_rectangle.init_from_rect(rectangle, PassIcon.BORDER_RADIUS)

        snapshot.push_rounded_clip(rounded_rectangle)
        snapshot.append_color(self.__background_color, rectangle)
        snapshot.pop()

    def __draw_icon(self, snapshot):
        if not self.__image:
            return

        rectangle = Graphene.Rect()
        rectangle.init(PassIcon.BORDER_WIDTH,
                       PassIcon.BORDER_WIDTH,
                       PassIcon.ICON_SIZE - PassIcon.BORDER_WIDTH,
                       PassIcon.ICON_SIZE - PassIcon.BORDER_WIDTH)

        rounded_rectangle = Gsk.RoundedRect()
        rounded_rectangle\
            .init_from_rect(rectangle,
                            PassIcon.BORDER_RADIUS - PassIcon.BORDER_WIDTH)

        texture = self.__image.as_texture()
        texture_height = texture.get_height()
        texture_width = texture.get_width()
        scale = (PassIcon.ICON_SIZE) / max(texture_width, texture_height)

        padding_top = (PassIcon.ICON_SIZE - (texture_height * scale)) / 2
        padding_left = (PassIcon.ICON_SIZE - (texture_width * scale)) / 2

        rect_texture = Graphene.Rect()
        rect_texture.init(padding_left + PassIcon.BORDER_WIDTH,
                          padding_top + PassIcon.BORDER_WIDTH,
                          texture_width * scale - PassIcon.BORDER_WIDTH,
                          texture_height * scale - PassIcon.BORDER_WIDTH)

        snapshot.push_rounded_clip(rounded_rectangle)
        snapshot.append_texture(texture, rect_texture)
        snapshot.pop()

    def __guess_background_color(self):
        if not self.__image:
            return

        pixel_buffer = self.__image.as_pixbuf()
        data = pixel_buffer.read_pixel_bytes().get_data()

        # This method assumes that the background color of an image is the color
        # of the first pixel of the image if it is not transparent.

        background_color = None
        if not pixel_buffer.get_has_alpha() or data[3] > 0:
            background_color = data[0:3]

        return Color(*background_color) if background_color else None

    def do_snapshot(self, snapshot):
        self.__draw_background(snapshot)
        self.__draw_icon(snapshot)

    def set_background_color(self, color):
        bg_color = self.__guess_background_color()
        if not bg_color:
            bg_color = color

        self.__background_color = bg_color.as_gdk_rgba()
        self.queue_draw()

    def set_image(self, image):
        self.__image = image
        self.queue_draw()




# pkpass.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, Gtk

from .digital_pass import Barcode, Color, Currency, Date, DigitalPass, Image, \
                          PassDataExtractor, PassField


class PKPass:
    """
    A representation of a PassKit pass
    """

    styles = ['boardingPass',
              'coupon',
              'eventTicket',
              'generic',
              'storeCard']

    __slots__ = ('__description', '__format_version', '__organization_name',
        '__pass_type_identifier', '__serial_number', '__team_identifier',
        '__expiration_date', '__voided', '__locations', '__maximum_distance',
        '__relevant_date', '__style', '__auxiliary_fields', '__back_fields',
        '__header_fields', '__primary_fields', '__secondary_fields',
        '__transit_type', '__barcode', '__barcodes', '__background',
        '__background_color', '__foreground_color', '__grouping_identifier',
        '__icon', '__label_color', '__logo', '__logo_text', '__strip',
        '__authentication_token', '__web_service_url')

    def __init__(self, pass_data, translation, images):
        data = PassDataExtractor(pass_data)

        self.__description = data.get('description')
        self.__format_version = data.get('formatVersion')
        self.__organization_name = data.get('organizationName')
        self.__pass_type_identifier = data.get('passTypeIdentifier')
        self.__serial_number = data.get('serialNumber')
        self.__team_identifier = data.get('teamIdentifier')

        self.__expiration_date = data.get('expirationDate', Date.from_iso_string)
        self.__voided = data.get('voided', bool)

        self.__locations = data.get('locations')
        self.__maximum_distance = data.get('maxDistance')
        self.__relevant_date = data.get('relevantDate', Date.from_iso_string)

        self.__style = None
        for style in PKPass.styles:
            if style in data.keys():
                self.__style = style
                break

        self.__auxiliary_fields = data.get(self.__style)\
            .get_list('auxiliaryFields', StandardField, translation)
        self.__back_fields = data.get(self.__style)\
            .get_list('backFields', StandardField, translation)
        self.__header_fields = data.get(self.__style)\
            .get_list('headerFields', StandardField, translation)
        self.__primary_fields = data.get(self.__style)\
            .get_list('primaryFields', StandardField, translation)
        self.__secondary_fields = data.get(self.__style)\
            .get_list('secondaryFields', StandardField, translation)
        self.__transit_type = data.get(self.__style).get('transitType')

        self.__barcode = data.get('barcode', Barcode)
        self.__barcodes = data.get_list('barcodes', Barcode)
        self.__background = Image(images['background']) \
            if 'background' in images else None
        self.__background_color = data.get('backgroundColor', Color.from_css)
        self.__foreground_color = data.get('foregroundColor', Color.from_css)
        self.__grouping_identifier = data.get('groupingIdentifier') \
            if self.style() in ['boardingPass', 'eventTicket'] else None
        self.__icon = Image(images['icon']) if 'icon' in images else None
        self.__label_color = data.get('labelColor', Color.from_css)
        self.__logo = Image(images['logo']) if 'logo' in images else None
        self.__logo_text = data.get('logoText')
        self.__strip = Image(images['strip']) if 'strip' in images else None

        self.__authentication_token = data.get('authenticationToken')
        self.__web_service_url = data.get('webServiceURL')


    # Standard

    # mandatory
    def description(self):
        return self.__description

    # mandatory
    def format_version(self):
        return self.__format_version

    # mandatory
    def organization_name(self):
        return self.__organization_name

    # mandatory
    def pass_type_identifier(self):
        return self.__pass_type_identifier

    # mandatory
    def serial_number(self):
        return self.__serial_number

    # mandatory
    def team_identifier(self):
        return self.__team_identifier


    # Expiration

    def expiration_date(self):
        return self.__expiration_date

    def voided(self):
        return self.__voided


    # Relevance

    def locations(self):
        return self.__locations

    def maximum_distance(self):
        return self.__maximum_distance

    def relevant_date(self):
        return self.__relevant_date


    # Style

    def style(self):
        return self.__style


    # Fields

    def auxiliary_fields(self):
        return self.__auxiliary_fields

    def back_fields(self):
        return self.__back_fields

    def header_fields(self):
        return self.__header_fields

    def primary_fields(self):
        return self.__primary_fields

    def secondary_fields(self):
        return self.__secondary_fields

    def transit_type(self):
        return self.__transit_type


    # Visual appearance

    def barcode(self):
        return self.__barcode

    def barcodes(self):
        return self.__barcodes

    def background(self):
        return self.__background

    def background_color(self):
        return self.__background_color

    def foreground_color(self):
        return self.__foreground_color

    def grouping_identifier(self):
        return self.__grouping_identifier

    def icon(self):
        return self.__icon

    def label_color(self):
        return self.__label_color

    def logo(self):
        return self.__logo

    def logo_text(self):
        return self.__logo_text

    def strip(self):
        return self.__strip

    # Web Service

    def authentication_token(self):
        return self.__authentication_token

    def web_service_url(self):
        return self.__web_service_url


class PKPassAdapter(DigitalPass):
    def __init__(self, pkpass):
        super().__init__()
        self.__adaptee = pkpass

    def adaptee(self):
        return self.__adaptee

    def additional_information(self):
        return self.__adaptee.back_fields()

    def background_color(self):
        return self.__adaptee.background_color()

    def barcodes(self):
        barcodes = self.__adaptee.barcodes()

        if not barcodes:
            barcodes = [self.__adaptee.barcode()]

        return barcodes

    def creator(self):
        return self.__adaptee.organization_name()

    def description(self):
        pass_style = self.__adaptee.style()
        fields = self.__adaptee.primary_fields()

        if pass_style == 'boardingPass' and len(fields) == 2:
            return '%s → %s' % (fields[0].label(), fields[1].label())

        return self.__adaptee.description()

    def expiration_date(self):
        return self.__adaptee.expiration_date()

    def file_extension():
        return '.pkpass'

    def format(self):
        return 'pkpass'

    def icon(self):
        return self.__adaptee.icon()

    def is_updatable(self):
        return self.__adaptee.web_service_url() and not self.has_expired()

    def mime_type():
        return 'application/vnd.apple.pkpass'

    def relevant_date(self):
        return self.__adaptee.relevant_date()

    def unique_identifier(self):
        return '.'.join([self.__adaptee.pass_type_identifier(),
                         self.__adaptee.serial_number(),
                         self.format()])

    def voided(self):
        return self.__adaptee.voided()


class StandardField(PassField):
    """
    A PKPass Standard Field
    """

    __slots__ = ('__key',
                 '__text_alignment')

    def __init__(self, pkpass_field_dictionary, translation_dictionary = None):
        self.__key = pkpass_field_dictionary['key']

        super().__init__()

        try:
            # Pass field values contain information, provided as a string, that
            # may have to be parsed and formatted. This is the case of numbers,
            # currencies and dates.

            value = pkpass_field_dictionary['value']

            if 'dateStyle' in pkpass_field_dictionary:
                value = Date.from_iso_string(value)

            elif 'currencyCode' in pkpass_field_dictionary:
                value = Currency.format(value, pkpass_field_dictionary['currencyCode'])

            elif translation_dictionary and value in translation_dictionary.keys():
                # The value is neither a date nor a currency
                value = translation_dictionary[value]

        except Exception:
            # If any error occur during the processing of the provided value,
            # this software will show the original text (as is). Because of
            # this, all exceptions produced in this block will be ignored.
            pass

        finally:
            self._value = value.strip()

        if not self.__key or not self._value:
            # Keys and values are required fields.
            raise Exception()

        if 'label' in pkpass_field_dictionary.keys():
            self._label = pkpass_field_dictionary['label']
            if translation_dictionary and self._label in translation_dictionary.keys():
                self._label = translation_dictionary[self._label]
            self._label = self._label.upper()

        self.__text_alignment = None
        if 'textAlignment' in pkpass_field_dictionary.keys():
            self.__text_alignment = pkpass_field_dictionary['textAlignment']

    def key(self):
        return self.__key

    def text_alignment(self):
        return self.__text_alignment



# main.py
#
# Copyright 2022-2024 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import gi

gi.require_version('Gdk', '4.0')
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from gi.repository import GLib, Gdk, Gio, Gtk, Adw

from .digital_pass import DigitalPass
from .digital_pass_factory import FileIsNotAPass, FormatNotSupportedYet, PassFactory
from .digital_pass_list_store import DigitalPassListStore
from .digital_pass_updater import PassUpdater
from .persistence import FileAlreadyImported, PersistenceManager
from .settings import Settings
from .window import PassesWindow


class Application(Adw.Application):

    # Application ID
    ID = 'me.sanchezrodriguez.passes'

    def __init__(self):
        super().__init__(application_id=Application.ID,
                         flags=Gio.ApplicationFlags.FLAGS_NONE)

        self.__file_chooser = None
        self.__persistence = PersistenceManager()
        self.__settings = Settings(Application.ID)
        self.__pass_list = DigitalPassListStore(self.__settings)

        pass_files = self.__persistence.load_pass_files()
        for pass_file in pass_files:
            digital_pass = PassFactory.create(pass_file)
            self.__pass_list.insert(digital_pass)

    def do_activate(self):
        window = self.props.active_window

        if not window:
            window = PassesWindow(application=self,
                                  pass_list_model=self.__pass_list)

        self.create_action('about', self.on_about_action)
        self.create_action('delete', self.on_delete_action)
        self.create_action('import', self.on_import_action, ['<Control>o'])
        self.create_action('quit', self.on_quit_action, ['<Control>q'])
        self.create_action('update', self.on_update_action, ['<Control>u'])

        pass_list_is_empty = self.__pass_list.is_empty()
        window.show_welcome_view(pass_list_is_empty)

        if not pass_list_is_empty:
            window.select_pass_at_index(0)
            window.show_pass_on_select(True)

        window.present()

    def do_startup(self):
        Adw.Application.do_startup(self)

    def import_pass(self, pass_file):
        try:
            digital_pass = PassFactory.create(pass_file)

            if digital_pass in self.__pass_list:
                self.window().show_toast("Pass already imported")
                return

            stored_file = self.__persistence\
                .save_pass_file(pass_file, digital_pass.unique_identifier())

            digital_pass.set_path(stored_file.get_path())
            self.__pass_list.insert(digital_pass)

            if self.window():
                if not self.__pass_list.is_empty():
                    self.window().show_welcome_view(False)

                found, index = self.__pass_list.find(digital_pass)
                if found:
                    self.window().show_pass_on_select(False)
                    self.window().select_pass_at_index(index)
                    self.window().show_pass_on_select(True)

        except Exception as exception:
            self.window().show_toast(str(exception))

    def on_about_action(self, widget, __):
        about = Adw.AboutDialog()
        about.set_application_icon('me.sanchezrodriguez.passes')
        about.set_application_name(_('Passes'))
        about.set_copyright('Copyright © 2022-2023 Pablo Sánchez Rodríguez')
        about.set_license_type(Gtk.License.GPL_3_0)
        about.set_developer_name('Pablo Sánchez Rodríguez')
        about.set_issue_url('https://github.com/pablo-s/passes/issues')
        about.set_version('0.10')
        about.set_website('https://github.com/pablo-s/passes')
        about.present(self.window())

    def on_delete_action(self, widget, _):
        if not self.window():
            return

        selected_pass = self.window().selected_pass()
        selected_pass_index = self.window().selected_pass_index()

        self.__persistence.delete_pass_file(selected_pass)
        self.__pass_list.remove(selected_pass_index)

        if self.__pass_list.is_empty():
            self.window().show_welcome_view(True)
            self.window().navigate_back()
            return

        index_to_select = min(self.__pass_list.length() - 1, selected_pass_index)
        self.window().show_pass_on_select(False)
        self.window().navigate_back()
        self.window().select_pass_at_index(index_to_select)
        self.window().show_pass_on_select(True)

    def on_import_action(self, widget, __):
        if not self.__file_chooser:
            self.__file_chooser = Gtk.FileDialog.new()

            supported_types_filter = Gtk.FileFilter()
            supported_types_filter.set_name(_('Supported passes'))
            for mime_type in DigitalPass.supported_mime_types():
                supported_types_filter.add_mime_type(mime_type)

            all_files_filter = Gtk.FileFilter()
            all_files_filter.set_name(_('All files'))
            all_files_filter.add_pattern('*')

            filter_list = Gio.ListStore.new(Gtk.FileFilter)
            filter_list.append(supported_types_filter)
            filter_list.append(all_files_filter)
            self.__file_chooser.set_filters(filter_list)

            self.__file_chooser.set_modal(True)

        self.__file_chooser.open(parent = self.window(),
                                 callback = self._on_file_chosen)

    def on_preferences_action(self, widget, _):
        print('app.preferences action activated')

    def on_quit_action(self, widget, _):
        self.window().close()

    def on_update_action(self, widget, __):
        """ Update currently selected pass """
        selected_pass = self.window().selected_pass()

        if not selected_pass:
            return

        try:
            # Download and save the latest version of the pass file
            latest_pass_data = PassUpdater.update(selected_pass)
            stored_file = self.__persistence\
                .save_pass_data(latest_pass_data,
                                selected_pass.unique_identifier() + '.tmp')

            # Create a new pass from the saved file
            digital_pass = PassFactory.create(stored_file)

            # Replace the old pass with the new one
            self.__pass_list.insert(digital_pass)
            selected_pass_index = self.window().selected_pass_index()
            self.__pass_list.remove(selected_pass_index)

            # Replace the old pass file with the new one
            self.__persistence.replace_pass_file(selected_pass,
                                                 replacement=digital_pass)

            # Select the new pass in the pass list
            found, updated_pass_index = self.__pass_list.find(digital_pass)
            self.window().select_pass_at_index(updated_pass_index)

            # Notify user
            self.window().show_toast(_('Pass updated'))

        except FileIsNotAPass:
            self.__persistence.delete_pass_file(stored_file)
            self.window().show_toast(_('Pass could not be updated'))

        except Exception as exception:
            self.window().show_toast(str(exception))

    def create_action(self, name, callback, shortcuts=None):
        """ Add an Action and connect to a callback """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)

        if shortcuts:
            self.set_accels_for_action(f'app.{name}',
                                       shortcuts)

    def _on_file_chosen(self, file_chooser, result):
        try:
            pass_file = file_chooser.open_finish(result)

            if not pass_file:
                return

            self.import_pass(pass_file)

        except Exception as exception:
            self.window().show_toast(exception.message)


    def window(self):
        return self.props.active_window

def main(version):
    app = Application()
    return app.run(sys.argv)



# persistence.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, tempfile

from gi.repository import Gio, GLib
from .digital_pass import DigitalPass


class PersistenceManager:
    """
    """

    def __init__(self):
        self.__data_dir = GLib.get_user_data_dir()
        self.__supported_file_extensions = DigitalPass.supported_file_extensions()

    def load_pass_files(self):
        file_names = os.listdir(self.__data_dir)
        pass_files = list()

        for file_name in file_names:
            basename, extension = os.path.splitext(file_name)

            if extension not in self.__supported_file_extensions:
                continue

            pass_file_path = os.path.join(self.__data_dir, file_name)
            pass_file = Gio.File.new_for_path(pass_file_path)
            pass_files.append(pass_file)

        return pass_files

    def delete_pass_file(self, a_pass):
        target_path = a_pass.get_path()
        target_file = Gio.File.new_for_path(target_path)
        target_file.delete()

    def replace_pass_file(self, pass_to_replace, replacement):
        source_path = replacement.get_path()
        destination_path = pass_to_replace.get_path()

        os.remove(destination_path)
        os.rename(source_path, destination_path)

        replacement.set_path(destination_path)

    def save_pass_data(self, pass_data, file_name):
        with tempfile.NamedTemporaryFile() as temp_pass_file:
            temp_pass_file.write(pass_data)

            pass_file_path = os.path.join(tempfile.gettempdir(), str(temp_pass_file.name))
            pass_file = Gio.File.new_for_path(pass_file_path)
            return self.save_pass_file(pass_file, file_name)

    def save_pass_file(self, pass_file, file_name):
        destination_file_path = os.path.join(self.__data_dir, file_name)
        destination_file = Gio.File.new_for_path(destination_file_path)

        if Gio.File.query_exists(destination_file):
            raise FileAlreadyImported()

        pass_file.copy(destination=destination_file,
                       flags=Gio.FileCopyFlags.NONE,
                       cancellable=None,
                       progress_callback=None,
                       progress_callback_data=None)

        return destination_file


class FileAlreadyImported(Exception):
    def __init__(self):
        message = _('File already imported')
        super().__init__(message)



# pkpass_row_header.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/pass_row_header.ui')
class PassRowHeader(Gtk.Label):

    __gtype_name__ = 'PassRowHeader'

    def __init__(self, text):
        super().__init__()
        self.set_text(str(text))



# window.py
#
# Copyright 2022-2024 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, GObject, Gtk, Adw

from .additional_information_pane import AdditionalInformationPane
from .barcode_dialog import BarcodeDialog
from .digital_pass_list_store import SortingCriteria
from .pass_list import PassList
from .pass_widget import PassWidget


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/window.ui')
class PassesWindow(Adw.ApplicationWindow):
    __gtype_name__ = 'PassesWindow'

    views = Gtk.Template.Child()

    main_split_view = Gtk.Template.Child()
    inner_split_view = Gtk.Template.Child()
    toast_overlay = Gtk.Template.Child()

    # Left panel
    pass_list = Gtk.Template.Child()

    # Main panel
    update_button = Gtk.Template.Child()
    info_button = Gtk.Template.Child()
    pass_widget = Gtk.Template.Child()

    # Right panel
    back_button  = Gtk.Template.Child()
    pass_additional_info  = Gtk.Template.Child()

    def __init__(self, pass_list_model, **kwargs):
        super().__init__(**kwargs)

        self.__show_pass_on_select = False

        # Set help overlay
        help_overlay = Gtk.Builder\
            .new_from_resource('/me/sanchezrodriguez/passes/help_overlay.ui')\
            .get_object('help_overlay')

        self.set_help_overlay(help_overlay)

        self.get_application()\
            .set_accels_for_action('win.show-help-overlay',
                                   ['<Control>question'])

        # Bind pass list and model
        self.pass_list.bind_model(pass_list_model)

        # Create action for pass sorting menu
        action = Gio.SimpleAction\
            .new_stateful("sort",
                          GLib.VariantType.new('s'),
                          GLib.Variant.new_string(pass_list_model.sorting_criteria()))

        action.connect("activate", self._on_sort_action)
        self.add_action(action)

        # Connect callbacks
        self.back_button.connect('clicked', self._on_back_button_clicked)
        self.pass_list.connect('pass-activated', self._on_pass_activated)
        self.pass_widget.connect('barcode-clicked', self._on_barcode_clicked)
        self.info_button.connect('clicked', self._on_info_button_clicked)

    def _on_back_button_clicked(self, button):
        self.inner_split_view.set_show_sidebar(False);

    def _on_barcode_clicked(self, button):
        try:
            selected_pass = self.selected_pass()
            barcode = selected_pass.barcodes()[0]

            if barcode:
                dialog = BarcodeDialog()
                dialog.set_barcode(barcode)
                dialog.present(self)

        except Exception as error:
            self.show_toast(str(error))

    def _on_info_button_clicked(self, button):
        self.inner_split_view.set_show_sidebar(True);

    def _on_pass_activated(self, pass_list, digital_pass):
        self.update_button.set_sensitive(digital_pass.is_updatable())

        self.pass_widget.content(digital_pass)
        self.pass_additional_info.content(digital_pass)

        if self.inner_split_view.get_collapsed():
            self.inner_split_view.set_show_sidebar(False);

        self.main_split_view.set_show_content(self.__show_pass_on_select)

    def _on_sort_action(self, action, target: GLib.Variant):
        sorting_criteria = SortingCriteria.from_string(target.get_string())
        self.pass_list.sort_by(sorting_criteria)
        action.set_state(target)

    def navigate_back(self):
        self.main_split_view.set_show_content(False)

    def select_pass_at_index(self, index):
        self.pass_list.select_pass_at_index(index)

    def selected_pass(self):
        return self.pass_list.selected_pass()

    def selected_pass_index(self):
        return self.pass_list.selected_pass_index()

    def show_pass_on_select(self, show_pass):
        self.__show_pass_on_select = show_pass

    def show_toast(self, message):
        toast = Adw.Toast.new(message)
        self.toast_overlay.add_toast(toast)

    def show_welcome_view(self, show=True):
        view_name = 'welcome' if show else 'main'
        self.views.set_visible_child_name(view_name)



# digital_pass_list_store.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from enum import StrEnum

from gi.repository import Gio, GObject

from .digital_pass import Date, DigitalPass


class DigitalPassListStore(GObject.GObject):

    __gtype_name__ = 'DigitalPassListStore'

    def __init__(self, settings):
        super().__init__()
        self.__settings = settings
        self.__list_store = Gio.ListStore.new(DigitalPass)
        self.__sorting_criteria = settings.get_sorting_criteria()

    def __contains__(self, digital_pass):
        return self.find(digital_pass)[0]

    def find(self, digital_pass):
        # The implementation of this method should use
        # Gio.ListStore.find_with_equal_func() instead of get_item(). However,
        # the method is broken and the fix has not been merged yet.
        # https://gitlab.gnome.org/GNOME/pygobject/-/merge_requests/218

        for position in range(self.length()):
            item = self.__list_store.get_item(position)
            if item.unique_identifier() == digital_pass.unique_identifier():
                return True, position

        return False, 0

    def get_model(self):
        return self.__list_store

    def insert(self, digital_pass):
        self.__list_store\
            .insert_sorted(digital_pass,
                           self.__sorting_criteria.sorting_function())

    def is_empty(self):
        return self.length() == 0

    def length(self):
        return len(self.__list_store)

    def remove(self, index):
        self.__list_store.remove(index)

    def sort_by(self, sorting_criteria):
        self.__sorting_criteria = sorting_criteria
        self.__list_store.sort(sorting_criteria.sorting_function())
        self.__settings.set_sorting_criteria(sorting_criteria)

    def sorting_criteria(self):
        return self.__sorting_criteria


class SortPassesBy:

    @classmethod
    def creator(cls, pass1, pass2):
        """
        Sort passes by creator.
        """
        return  pass1.creator().lower() > pass2.creator().lower()

    @classmethod
    def description(cls, pass1, pass2):
        """
        Sort passes by description.
        """
        return  pass1.description().lower() > pass2.description().lower()

    @classmethod
    def expiration_date(cls, d1, d2):
        """
        Sort passes by expiration date. In the event that two passes have the
        same expiration date then they will be sorted by description.
        """
        dates_comparison = Date.compare_dates(d1.expiration_date(),
                                              d2.expiration_date())

        d1_is_later_than_d2 = dates_comparison > 0
        dates_are_equal = dates_comparison == 0

        return  d1_is_later_than_d2 or \
                (dates_are_equal and d1.description() > d2.description())

    @classmethod
    def relevant_date(cls, pass1, pass2):
        """
        Sort passes by relevant date. In the event that two passes have the
        same expiration date then they will be sorted by description.
        """
        dates_comparison = Date.compare_dates(pass1.relevant_date(),
                                              pass2.relevant_date())

        d1_is_later_than_d2 = dates_comparison > 0
        dates_are_equal = dates_comparison == 0

        return  d1_is_later_than_d2 or \
                (dates_are_equal and pass1.description() > pass2.description())


class SortingCriteria(StrEnum):
    CREATOR = 'creator'
    DESCRIPTION = 'description'
    EXPIRATION_DATE = 'expiration_date'
    RELEVANT_DATE = 'relevant_date'

    __SORTING_FUNCTIONS = {
        CREATOR: SortPassesBy.creator,
        DESCRIPTION: SortPassesBy.description,
        EXPIRATION_DATE: SortPassesBy.expiration_date,
        RELEVANT_DATE: SortPassesBy.relevant_date
    }

    @classmethod
    def from_string(cls, criteria):
        return cls[criteria.upper()]

    def sorting_function(self):
        return self.__SORTING_FUNCTIONS[self.value]



# digital_pass_factory.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import re
import zipfile

from gi.repository import Gdk, GObject, Gtk

from .espass import EsPass, EsPassAdapter
from .pkpass import PKPass, PKPassAdapter


def decode_string(string):
    encodings = ['utf-8', 'utf-16']
    decoded_string = ''

    for encoding in encodings:
        try:
            decoded_string = string.decode(encoding)
            break
        except UnicodeDecodeError:
            pass

    if not decoded_string:
        raise UnknownEncoding()

    return decoded_string


class PassFactory:
    """
    Create a digital pass
    """

    @classmethod
    def create(cls, pass_file):
        try:
            path = pass_file.get_path()
            archive = zipfile.ZipFile(path, 'r')

            if 'main.json' in archive.namelist():
                digital_pass = cls.__create_espass(archive)
            elif 'pass.json' in archive.namelist():
                digital_pass = cls.__create_pkpass(archive)
            else:
                raise FileIsNotAPass()

            digital_pass.set_path(path)
            return digital_pass

        except zipfile.BadZipFile as exception:
            raise FileIsNotAPass()

    @classmethod
    def __create_espass(cls, archive):
        """
        Create an EsPass object from a compressed file
        """

        pass_data = dict()
        pass_images = dict()

        for file_name in archive.namelist():
            if file_name.endswith('.png'):
                image = archive.read(file_name)
                pass_images[file_name] = image

            if file_name.endswith('main.json'):
                json_content = archive.read(file_name)
                pass_data = json.loads(json_content)

        espass = EsPass(pass_data, pass_images)
        return EsPassAdapter(espass)

    @classmethod
    def __create_pkpass(cls, archive):
        """
        Create a PKPass object from a compressed file
        """

        manifest_text = archive.read('manifest.json')
        manifest = json.loads(manifest_text)

        pass_data = dict()
        pass_translations = dict()
        pass_images = dict()

        for file_name in sorted(manifest.keys()):
            if file_name.endswith('.png'):
                image = archive.read(file_name)

                # For every type of image (background, footer, icon, logo, strip
                # and thumbnail), only load the image with lowest resolution

                image_type = re.split('\.|@', file_name)[0]

                if image_type in pass_images.keys():
                    continue

                pass_images[image_type] = image

            if file_name.endswith('pass.strings'):
                language = file_name.split('.')[0]
                file_content = archive.read(file_name)
                translation_dict = cls.__create_translation_dict(file_content)
                pass_translations[language] = translation_dict

            if file_name.endswith('pass.json'):
                json_content = archive.read(file_name)
                pass_data = json.loads(json_content)


        language_to_import = None
        if pass_translations:
            user_language = Gtk.get_default_language().to_string()

            for language in pass_translations:
                if language in user_language:
                    language_to_import = language
                    break

            if language_to_import is None:
                # TODO: Open a dialogue and ask the user what language to import
                pass

        pass_translation = None
        if language_to_import:
            pass_translation = pass_translations[language_to_import]

        pkpass = PKPass(pass_data, pass_translation, pass_images)
        return PKPassAdapter(pkpass)

    @classmethod
    def __create_translation_dict(cls, translation_file_content):
        content = decode_string(translation_file_content)
        entries = content.split(';')

        translation_dict = dict()

        for entry in entries:
            result = re.search('"(.*)" = "(.*)"', entry, re.DOTALL)

            if not result or len(result.groups()) != 2:
                continue

            translation_key = result.group(1)
            translation_value = result.group(2)
            translation_dict[translation_key] = translation_value

        return translation_dict


class FileIsNotAPass(Exception):
    def __init__(self):
        message = _('File is not a pass')
        super().__init__(message)


class FormatNotSupportedYet(Exception):
    def __init__(self):
        message = _('Format not supported yet')
        super().__init__(message)


class UnknownEncoding(Exception):
    def __init__(self):
        message = _('Unknown file encoding')
        super().__init__(message)



# digital_pass_updater.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import http.client


class PassKitWebService:

    @staticmethod
    def get_latest_version(web_service_url, pass_type_identifier,
                           serial_number, authentication_token):

        endpoint = '{}/v1/passes/{}/{}'.format(web_service_url,
                                               pass_type_identifier,
                                               serial_number)

        authorization = 'ApplePass {}'.format(authentication_token)
        headers = {'Authorization' : authorization}

        protocol, host_and_path = endpoint.split('//', 1)
        host, path = host_and_path.split('/', 1)

        connection = http.client.HTTPSConnection(host)
        connection.request("GET", '/{}'.format(path), headers=headers)

        return connection

    @staticmethod
    def get_from_new_location(new_location):
        protocol, host_and_path = new_location.split('//', 1)
        host, path = host_and_path.split('/', 1)

        connection = http.client.HTTPSConnection(host)
        connection.request("GET", '/{}'.format(path))

        return connection

class PassUpdater:

    @classmethod
    def update(this_class, a_pass):
        """ Download and return the latest version of a digital pass """

        if not a_pass.is_updatable() or a_pass.format() != 'pkpass':
            raise PassNotUpdatable()

        return this_class._update_pkpass(a_pass.adaptee())

    @classmethod
    def _update_pkpass(this_class, pkpass):
        """ Download and return the latest version of a PKPass """

        web_service_url = pkpass.web_service_url()
        pass_type_identifier = pkpass.pass_type_identifier()
        serial_number = pkpass.serial_number()
        authentication_token = pkpass.authentication_token()

        first_iteration = True
        last_response_status = -1
        new_location = None

        while first_iteration or last_response_status in [301, 302]:
            first_iteration = False

            if new_location:
                connection = PassKitWebService\
                                .get_from_new_location(new_location)
            else:
                connection = PassKitWebService\
                                .get_latest_version(web_service_url,
                                                    pass_type_identifier,
                                                    serial_number,
                                                    authentication_token)

            response = connection.getresponse()
            last_response_status = response.status

            if response.status == 200:
                """ 200 OK """
                pkpass_data = response.read()
                connection.close()
                return pkpass_data

            elif response.status in [204, 304]:
                """ 204 No Content / 304 Not Modified """
                connection.close()
                raise PassAlreadyUpdated()

            elif response.status in [301, 302]:
                """ 301 Moved Permanently / 302 Found """
                # Web Service URL has been updated
                new_location = response.getheader('Location')
                connection.close()
                continue

            raise PassUpdateError(response.status, response.reason)


class PassAlreadyUpdated(Exception):
    def __init__(self):
        message = _('Pass already updated')
        super().__init__(message)


class PassNotUpdatable(Exception):
    def __init__(self):
        message = _('Pass not updatable')
        super().__init__(message)


class PassUpdateError(Exception):
    def __init__(self, error_code, reason):
        message = _('Pass update error: {} {}').format(error_code, reason)
        super().__init__(message)



# barcode_widget.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Adw, Gdk, Graphene, Gsk, Gtk

from .barcode_content_encoder import BarcodeContentEncoder
from .digital_pass import Color


class BarcodeWidget(Gtk.Widget):

    __gtype_name__ = 'BarcodeWidget'

    FOREGROUND = '1';
    BACKGROUND = '2';

    def __init__(self):
        super().__init__()

        self.__data = []
        self.__data_width = 0
        self.__data_height = 0

        # Set background color to white
        self.__background_color = Color.named('white').as_gdk_rgba()

        # Set foreground color to black
        self.__foreground_color = Color.named('black').as_gdk_rgba()

        # Amount of barcode dots/modules that should fit in every margin (either
        # horizontal or vertical)
        self.__margin_size = 4

    def aspect_ratio(self):
        return (self.__data_width + self.__margin_size) / (self.__data_height + self.__margin_size)

    def do_snapshot(self, snapshot):
        canvas_width = self.get_allocated_width()
        canvas_height = self.get_allocated_height()

        barcode_width = self.__data_width + 2 * self.__margin_size
        barcode_height = self.__data_height + 2 * self.__margin_size

        h_scaling = canvas_width / barcode_width
        v_scaling = canvas_height / barcode_height
        scaling_factor = int(min(h_scaling, v_scaling))

        translation = Graphene.Point()
        translation.x = (canvas_width - barcode_width * scaling_factor) / 2
        translation.y = (canvas_height - barcode_height * scaling_factor) / 2
        snapshot.translate(translation)

        snapshot.scale(scaling_factor, scaling_factor)


        # Draw the barcode
        for i in range(self.__data_height):
            for j in range(self.__data_width):

                is_foreground = \
                    self.__data[i * self.__data_width + j] == BarcodeWidget.FOREGROUND

                if not is_foreground:
                    continue

                rectangle = Graphene.Rect()
                rectangle.init(j + self.__margin_size,
                               i + self.__margin_size,
                               1,
                               1)
                snapshot.append_color(self.__foreground_color, rectangle)

    def encode(self, format, message, encoding):
        encoding_function = None

        if format in ['AZTEC', 'PKBarcodeFormatAztec']:
            encoding_function = BarcodeContentEncoder.encode_aztec_code
            self.__margin_size = 2

        elif format in ['CODE_128', 'PKBarcodeFormatCode128']:
            encoding_function = BarcodeContentEncoder.encode_code128_code
            self.__margin_size = 7

        elif format in ['PDF_417', 'PKBarcodeFormatPDF417']:
            encoding_function = BarcodeContentEncoder.encode_pdf417_code
            self.__margin_size = 2

        elif format in ['PKBarcodeFormatQR', 'QR_CODE']:
            encoding_function = BarcodeContentEncoder.encode_qr_code

        else:
            raise BarcodeFormatNotSupported()

        module_list, width, height = encoding_function(message, encoding)

        self.__data = module_list
        self.__data_width = width
        self.__data_height = height

    def minimum_height(self):
        return self.__data_height + 2 * self.__margin_size

    def minimum_width(self):
        return self.__data_width + 2 * self.__margin_size


class BarcodeFormatNotSupported(Exception):
    def __init__(self):
        message = _('Barcode format not supported')
        super().__init__(message)



# barcode_dialog.py
#
# Copyright 2022-2024 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Adw, Gtk


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/barcode_dialog.ui')
class BarcodeDialog(Adw.Dialog):

    __gtype_name__ = 'BarcodeDialog'

    barcode = Gtk.Template.Child()
    alternative_text = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

    def set_barcode(self, code):

        self.barcode\
            .encode(code.format(), code.message(), code.message_encoding())

        alternative_text = code.alternative_text()
        if alternative_text:
            self.alternative_text.set_text(alternative_text)



# digital_pass.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import locale
import re

from gi.repository import Gdk, GdkPixbuf, GLib, GObject


class DigitalPass(GObject.GObject):

    __gtype_name__ = 'DigitalPass'

    def __init__(self):
        super().__init__()
        self.__path = None

    def additional_information(self):
        raise NotImplementedError()

    def background_color(self):
        raise NotImplementedError()

    def barcodes(self):
        raise NotImplementedError()

    def creator(self):
        raise NotImplementedError()

    def description(self):
        raise NotImplementedError()

    def expiration_date(self):
        raise NotImplementedError()

    def file_extension(self):
        raise NotImplementedError()

    def format(self):
        raise NotImplementedError()

    def get_path(self):
        return self.__path

    def has_expired(self):
        expiration_date = self.expiration_date()
        return (expiration_date and Date.now() > expiration_date) \
                or self.voided()

    def icon(self):
        raise NotImplementedError()

    def is_updatable(self):
        raise NotImplementedError()

    def mime_type():
        raise NotImplementedError()

    def relevant_date():
        raise NotImplementedError()

    def set_path(self, new_path: str):
        self.__path = new_path

    def unique_identifier(self):
        raise NotImplementedError()

    def voided(self):
        raise NotImplementedError()

    @classmethod
    def supported_mime_types(cls):
        return [pass_type.mime_type() for pass_type in cls.__subclasses__()]

    @classmethod
    def supported_file_extensions(cls):
        return [pass_type.file_extension() for pass_type in cls.__subclasses__()]


class Barcode:

    def __init__(self, barcode_dictionary):
        self.__format = barcode_dictionary['format']
        if not self.__format:
            # Format is a required field
            raise Exception()

        self.__message = barcode_dictionary['message']
        if not self.__message:
            # Message is a required field
            raise Exception()

        self.__message_encoding = None
        if 'messageEncoding' in barcode_dictionary.keys():
            self.__message_encoding = barcode_dictionary['messageEncoding']

        self.__alt_text = None
        if 'altText' in barcode_dictionary.keys():
            self.__alt_text = barcode_dictionary['altText']

    def alternative_text(self):
        return self.__alt_text

    def format(self):
        return self.__format

    def message(self):
        return self.__message

    def message_encoding(self):
        return self.__message_encoding


class Color:

    def __init__(self, r, g, b, a = 255):
        self.__r = int(r)
        self.__g = int(g)
        self.__b = int(b)
        self.__a = int(a)

    def red(self):
        return self.__r

    def green(self):
        return self.__g

    def blue(self):
        return self.__b

    def as_gdk_rgba(self):
        rgba = Gdk.RGBA()
        rgba.red = self.__r / 255
        rgba.green = self.__g / 255
        rgba.blue = self.__b / 255
        rgba.alpha = self.__a / 255
        return rgba

    def as_tuple(self):
        return (self.__r, self.__g, self.__b)

    @classmethod
    def from_css(this_class, css_string):
        if css_string.startswith('rgb'):
            result = re.search('rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)',
                               css_string)

            if not result or len(result.groups()) != 3:
                raise BadColor()

            r = result.group(1)
            g = result.group(2)
            b = result.group(3)

        elif css_string.startswith('#'):
            result = re.search('\#(\S{2})(\S{2})(\S{2})(\S{2})',
                               css_string)

            if not result or len(result.groups()) != 4:
                raise BadColor()

            r = int(result.group(2), 16)
            g = int(result.group(3), 16)
            b = int(result.group(4), 16)

        else:
            raise BadColor()

        return Color(r, g, b)

    def invert(self):
        self.__r = 255 - self.__r
        self.__g = 255 - self.__g
        self.__b = 255 - self.__b

    @classmethod
    def named(cls, color_name):
        if color_name == 'black':
            return Color(0, 0, 0, 255)
        elif color_name == 'white':
            return Color(255, 255, 255, 255)
        else:
            raise BadColor()


class Currency:

    SYMBOLS = {'CNY': '¥‎', 'EUR': '€', 'GBP': '£', 'INR': '₹', 'JPY': '¥',
               'KRW': '₩', 'RUB': '₽‎', 'USD': '$'}

    @classmethod
    def format(cls, amount, international_code):
        symbol_to_show = cls.get_symbol_from_code(international_code)
        output = locale.currency(amount, symbol=symbol_to_show, grouping=True)

        if symbol_to_show:
            localeconv = locale.localeconv()
            symbol_to_replace = localeconv['currency_symbol']

            if symbol_to_replace != symbol_to_show:
                output = output.replace(symbol_to_replace, symbol_to_show)

        return output

    @classmethod
    def get_symbol_from_code(cls, code):
        localeconv = locale.localeconv()
        local_currency_code = localeconv['int_curr_symbol']

        if code == local_currency_code:
            return localeconv['currency_symbol']

        if code in cls.SYMBOLS:
            return cls.SYMBOLS[code]

        return None


class Date:

    days_of_the_week = (_('Monday'), _('Tuesday'), _('Wednesday'),
                        _('Thursday'), _('Friday'), _('Saturday'), _('Sunday'))

    MAX = GLib.DateTime.new_utc(9999, 12, 31, 23, 59, 59)
    MIN = GLib.DateTime.new_utc(1, 1, 1, 0, 0, 0)

    def __init__(self, date):
        self.__date = date

    def __eq__(self, other):
        return self.compare(other) == 0

    def __gt__(self, other):
        return self.compare(other) > 0

    def __lt__(self, other):
        return self.compare(other) < 0

    def __str__(self):
        return self.__date.to_local().format('%c')

    def as_relative_pretty_string(self):

        now = GLib.DateTime.new_now_utc()
        today = GLib.Date.new_dmy(now.get_day_of_month(),
                                  now.get_month(),
                                  now.get_year())

        this = GLib.Date.new_dmy(self.__date.get_day_of_month(),
                                 self.__date.get_month(),
                                 self.__date.get_year())

        difference_in_days = GLib.Date.days_between(today, this)

        if difference_in_days == 0:
            return _('Today')

        if difference_in_days == 1:
            return _('Tomorrow')

        if 0 < difference_in_days < 7:
            return Date.days_of_the_week[self.__date.get_day_of_week() - 1]

        return self.__date.to_local().format('%x')

    def compare(self, other):
        return self.__date.compare(other.__date)

    @classmethod
    def compare_dates(cls, date1, date2):
        if not date1 and not date2:
            return 0

        if date1 and not date2:
            return -1

        if not date1 and date2:
            return 1

        return date1.compare(date2)

    @classmethod
    def from_iso_string(cls, string):
        """
        Creates a Date corresponding to the given ISO 8601 formatted string
        """

        # Include hours and seconds if they have not been specified.
        # Why? DateTime.new_from_iso8601 does not accept times without them.

        matches = re.finditer('(T|t)([0-9]{2}\:?)+(\+|\-|Z)?', string)

        for match in matches:
            missing_info = None
            colon_count = match.group(0).count(':')

            if colon_count == 0:
                # Minutes and seconds are missing
                missing_info = ':00:00'
            elif colon_count == 1:
                # Seconds are missing
                missing_info = ':00'

            if missing_info:
                insertion_index = match.end() - 1
                string = string[:insertion_index] + missing_info + string[insertion_index:]

            # We are only interested in the first match
            break

        date = GLib.DateTime.new_from_iso8601(string)
        return Date(date)

    @classmethod
    def now(cls):
        date = GLib.DateTime.new_now_local()
        return Date(date)


class Image:

    def __init__(self, image_data):
        self.__data = image_data

    def as_pixbuf(self):
        loader = GdkPixbuf.PixbufLoader()
        loader.write(self.__data)
        loader.close()
        return loader.get_pixbuf()

    def as_texture(self):
        return Gdk.Texture.new_from_bytes(GLib.Bytes(self.__data))


class PassDataExtractor:
    """
    A PassDataExtractor contains a reference to a dictionary and a set of helper
    functions to easy the process of creating a pass.
    """

    def __init__(self, dictionary):
        self._dictionary = dictionary

    def _cast_to_boolean(self, value):
        """
        Protected method that creates a boolean from a string.
        """
        return True if value.lower().startswith('true') else False

    def get(self, key, type_constructor=None):
        """
        Return an element from the dictionary using the provided key.

        If a constructor is specified, it will be used to create the instance
        that will be returned.
        """
        try:
            value = None

            if key in self._dictionary:
                value = self._dictionary[key]
            else:
                # The key does not exist... nothing to do
                return None

            if not type_constructor and type(value) == dict:
                return PassDataExtractor(value)

            if type_constructor:
                if type(type_constructor) == bool:
                    value = self._cast_to_boolean(value)

                else:
                    value = type_constructor(value)

            return value

        except:
            return None

    def get_list(self, key, item_constructor=None, extra_arguments=None):
        """
        Return a list of elements from the dictionary using the provided key.

        If a constructor is specified, it will be used to create each of the
        items that will be appended to the list.
        """

        data_list = self.get(key)

        if not data_list:
            return []

        if not extra_arguments:
            extra_arguments = ()

        elif type(extra_arguments) != tuple:
            extra_arguments = (extra_arguments,)

        result = list()
        for item_data in data_list:

            if item_constructor:
                try:
                    arguments = (item_data,) + extra_arguments
                    instance = item_constructor(*arguments)
                except:
                    continue
            else:
                instance = item_data

            result.append(instance)

        return result

    def keys(self):
        """
        Return the set of keys
        """
        return self._dictionary.keys()


class PassField:

    __slots__ = ('_label',
                 '_value')

    def __init__(self):
        self._label = None
        self._value = None

    def label(self):
        return self._label

    def value(self):
        return self._value


class TimeInterval:
    def __init__(self, start_time, end_time):
        self.__start_time = start_time
        self.__end_time = end_time

    def __contains__(self, time):
        if (self.__start_time and time < self.__start_time) or \
           (self.__end_time and time > self.__end_time):
            return False

        return True

    @classmethod
    def from_iso_strings(cls, start_time, end_time):
        start_time = Date.from_iso_string(start_time) if start_time else Date.MIN
        end_time = Date.from_iso_string(end_time) if end_time else Date.MAX
        return TimeInterval(start_time, end_time)

    def end_time(self):
        return self.__end_time


class BadColor(Exception):
    pass



# pass_list.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk

from .digital_pass import DigitalPass
from .digital_pass_list_store import SortingCriteria
from .pass_row import PassRow


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/pass_list.ui')
class PassList(Gtk.ListBox):

    __gtype_name__ = 'PassList'

    __gsignals__ = {
        'pass-activated' : (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE, (DigitalPass,))
    }

    def __init__(self):
        super().__init__()

        self.__list_model = None
        self.__selected_row = None
        self.__sorting_criteria = None
        self.set_header_func(self._on_update_headers)

        self.connect('row-activated', self._on_row_activated)

    def _on_row_activated(self, pass_list, pass_row):
        self.__selected_row = pass_row
        self.emit('pass-activated', pass_row.data())

    def _on_update_headers(self, row, row_above):
        row.update_header_text_for(self.__sorting_criteria)

        if row_above:
            row_above.update_header_text_for(self.__sorting_criteria)

        if not row_above or row.header_text() != row_above.header_text():
            row.show_header()
        else:
            row.hide_header()

    def bind_model(self, pass_list_model):
        self.__sorting_criteria = pass_list_model.sorting_criteria()
        self.__list_model = pass_list_model
        super().bind_model(pass_list_model.get_model(), PassRow)

    def row_at_index(self, index):
        row_at_index = self.get_row_at_index(index)

        if not row_at_index:
            row_at_index = self.get_row_at_index(0)

        return row_at_index if row_at_index else None

    def select_pass_at_index(self, index):
        row_at_index = self.row_at_index(index)

        if row_at_index:
            row_at_index.activate()

    def selected_pass(self):
        selected_pass = None

        if self.__selected_row:
            selected_pass = self.__selected_row.data()

        return selected_pass

    def selected_pass_index(self):
        index = None

        if self.__selected_row:
            index = self.__selected_row.get_index()

        return index

    def sort_by(self, sorting_criteria):
        self.__sorting_criteria = sorting_criteria
        self.__list_model.sort_by(sorting_criteria)




# pass_widget.py
#
# Copyright 2022-2024 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Adw, Gdk, GObject, Graphene, Gsk, Gtk, Pango

from .barcode_widget import BarcodeWidget
from .digital_pass import Color


PASS_WIDTH = 320
PASS_HEIGHT = 420
PASS_MARGIN = 12
BACKGROUND_BLUR_RADIUS = 30


class PassFont:
    label = Pango.FontDescription.new()
    label.set_size(9 * Pango.SCALE)
    label.set_weight(600)

    value = Pango.FontDescription.new()
    value.set_size(11 * Pango.SCALE)

    big_value = Pango.FontDescription.new()
    big_value.set_size(17 * Pango.SCALE)

    biggest_value = Pango.FontDescription.new()
    biggest_value.set_size(24 * Pango.SCALE)


class FieldLayout:

    def __init__(self,
                 pango_context,
                 field,
                 label_font = PassFont.label,
                 value_font = PassFont.value,
                 alignment = Pango.Alignment.LEFT):

        self.__label = Pango.Layout(pango_context)
        self.__label.set_alignment(alignment)
        self.__label.set_font_description(label_font)
        self.__label.set_text(field.label() if field.label() else '')

        self.__value = Pango.Layout(pango_context)
        self.__value.set_alignment(alignment)
        self.__value.set_font_description(value_font)
        self.__value.set_text(str(field.value()))
        self.__value.set_wrap(Pango.WrapMode.WORD_CHAR)

        width = max(self.__label.get_pixel_size().width,
                    self.__value.get_pixel_size().width)

        width = min(width, PASS_WIDTH - 2 * PASS_MARGIN) * Pango.SCALE

        self.__value.set_width(width)
        self.__label.set_width(width)

    def append(self, snapshot, label_color, value_color):
        label_height = self.__label.get_pixel_size().height
        value_height = self.__value.get_pixel_size().height

        snapshot.save()
        snapshot.append_layout(self.__label, label_color)

        point = Graphene.Point()
        point.y = label_height
        snapshot.translate(point)

        snapshot.append_layout(self.__value, value_color)
        snapshot.restore()

    def get_height(self):
        return self.__label.get_pixel_size().height + self.__value.get_pixel_size().height

    def get_width(self):
        return self.__label.get_width() / Pango.SCALE

    def set_alignment(self, alignment):
        self.__label.set_alignment(alignment)
        self.__value.set_alignment(alignment)

    def set_width(self, width):
        self.__label.set_width(width * Pango.SCALE)
        self.__value.set_width(width * Pango.SCALE)


class PassPlotter:

    def __init__(self, a_pass, pass_widget):
        self._pass_widget = pass_widget
        self._pango_context = pass_widget.get_pango_context()

    def _create_fields_layouts(self, fields):
        rows = []
        spacing_per_row = []

        current_row = []
        accumulated_width = 0

        max_row_width = PASS_WIDTH - 2 * PASS_MARGIN

        for field in fields:
            field_layout = FieldLayout(self._pango_context, field)
            field_width = field_layout.get_width()

            if (accumulated_width + field_width) + len(current_row) * PASS_MARGIN < max_row_width:
                current_row.append(field_layout)
                accumulated_width += field_width
                continue

            spacing = (max_row_width - accumulated_width) / (len(current_row)-1) if len(current_row) > 1 else 0
            spacing_per_row.append(spacing)
            rows.append(current_row)

            accumulated_width = field_width
            current_row = []
            current_row.append(field_layout)

        if current_row:
            spacing = (max_row_width - accumulated_width) / (len(current_row)-1) if len(current_row) > 1 else 0
            spacing_per_row.append(spacing)
            rows.append(current_row)

        return rows, spacing_per_row

    def _plot_background(self):
        rectangle = Graphene.Rect()
        rectangle.init(0, 0, PASS_WIDTH, PASS_HEIGHT)
        self._snapshot.append_color(self._bg_color, rectangle)

    def _plot_fields_layouts(self, fields):
        self._snapshot.save()

        point = Graphene.Point()
        point.x = PASS_MARGIN
        point.y = 0
        self._snapshot.translate(point)

        row_height = 0
        rows, spacing_per_row = self._create_fields_layouts(fields)

        for row in rows:
            row_height = 0
            spacing = spacing_per_row.pop(0)
            amount_of_fields = len(row)

            self._snapshot.save()
            for index, field_layout in enumerate(row):

                # Decide the alignment of the label and value according to the
                # location of the field in the row.

                if index == 0:
                    field_layout.set_alignment(Pango.Alignment.LEFT)
                elif index == amount_of_fields - 1:
                    field_layout.set_alignment(Pango.Alignment.RIGHT)
                else:
                    field_layout.set_alignment(Pango.Alignment.CENTER)

                # Plot the standard field
                field_layout.append(self._snapshot, self._label_color, self._fg_color)

                layout_height = field_layout.get_height()
                if layout_height > row_height:
                    row_height = layout_height

                # Add a horizontal space between fields
                point.x = field_layout.get_width() + spacing
                point.y = 0
                self._snapshot.translate(point)

            self._snapshot.restore()

            # Add a vertical space between rows
            point.x = 0
            point.y = row_height + 6
            self._snapshot.translate(point)

        self._snapshot.restore()

        # Perform a translation so that the next drawing starts below this one
        point.x = 0
        point.y = row_height * len(rows) + PASS_MARGIN
        self._snapshot.translate(point)

    @classmethod
    def new(clss, a_pass, pass_widget):
        if a_pass.format() == 'pkpass':
            return PkPassPlotter.new(a_pass, pass_widget)
        return EsPassPlotter(a_pass, pass_widget)

    def plot(self, snapshot):
        raise NotImplementedError()


class EsPassPlotter(PassPlotter):

    def __init__(self, a_pass, pass_widget):
        super().__init__(a_pass, pass_widget)
        espass = a_pass.adaptee()

        # Accent color
        accent_color = espass.accent_color()
        self._accent_color = accent_color.as_gdk_rgba() \
            if accent_color else Gdk.RGBA()

        # Background color
        self._bg_color = Color.named('white').as_gdk_rgba()

        # Foreground color
        self._fg_color = Color.named('black').as_gdk_rgba()

        # Label color
        self._label_color = self._fg_color.copy()

        # Logo
        self._logo_texture  = None

        if espass.icon():
            self._logo_texture  = espass.icon().as_texture()

        # Fields
        self._fields = espass.front_fields()

    def _plot_background(self):
        rectangle = Graphene.Rect()
        rectangle.init(0, 0, PASS_WIDTH, PASS_HEIGHT)
        self._snapshot.append_color(self._bg_color, rectangle)

    def _plot_fields(self):
        self._plot_fields_layouts(self._fields)

    def _plot_header(self):
        header_height = 32

        rectangle = Graphene.Rect()
        rectangle.init(0, 0, PASS_WIDTH, header_height + 2 * PASS_MARGIN)
        self._snapshot.append_color(self._accent_color, rectangle)

        # Draw the logo if it exists
        if self._logo_texture:
            logo_scale = header_height / self._logo_texture.get_height()
            logo_width = self._logo_texture.get_width() * logo_scale

            rectangle = Graphene.Rect()
            rectangle.init(PASS_MARGIN, PASS_MARGIN, logo_width, header_height)
            self._snapshot.append_texture(self._logo_texture, rectangle)

        # Perform a translation so that the next drawing starts below this one
        point = Graphene.Point()
        point.y = header_height + 3 * PASS_MARGIN
        self._snapshot.translate(point)

    def plot(self, snapshot):
        self._snapshot = snapshot

        self._snapshot.save()
        self._plot_background()
        self._plot_header()
        self._plot_fields()
        self._snapshot.restore()


class PkPassPlotter(PassPlotter):

    PRIMARY_FIELD_LABEL_FONT = PassFont.label
    PRIMARY_FIELD_VALUE_FONT = PassFont.biggest_value

    def __init__(self, a_pass, pass_widget):
        super().__init__(a_pass, pass_widget)

        # At this point we know we are going to plot a PKPass
        pkpass = a_pass.adaptee()

        # Background color
        bg_color = pkpass.background_color()
        self._bg_color = bg_color.as_gdk_rgba() \
            if bg_color else Color.named('white').as_gdk_rgba()

        # Foreground color
        fg_color = pkpass.foreground_color()
        self._fg_color = fg_color.as_gdk_rgba() \
            if fg_color else Color.named('black').as_gdk_rgba()

        # Label color
        label_color = pkpass.label_color()
        self._label_color = label_color.as_gdk_rgba() \
            if label_color else Color.named('black').as_gdk_rgba()

        # Images
        self._background_texture = None
        self._logo_texture = None
        self._strip_texture = None

        if pkpass.background():
            self._background_texture = pkpass.background().as_texture()

        if pkpass.logo():
            self._logo_texture = pkpass.logo().as_texture()

        if pkpass.strip():
            self._strip_texture = pkpass.strip().as_texture()

        # Fields
        self._header_fields = pkpass.header_fields()
        self._primary_fields = pkpass.primary_fields()
        self._secondary_fields = pkpass.secondary_fields()
        self._auxiliary_fields = pkpass.auxiliary_fields()

    @classmethod
    def new(clss, a_pass, pass_widget):
        pkpass = a_pass.adaptee()
        style = pkpass.style()

        if style == 'boardingPass':
            return BoardingPassPlotter(a_pass, pass_widget)
        elif style in ['coupon', 'storeCard']:
            return CouponPlotter(a_pass, pass_widget)
        elif style == 'eventTicket':
            return EventTicketPlotter(a_pass, pass_widget)
        elif style == 'generic':
            return GenericPlotter(a_pass, pass_widget)

    def plot(self, snapshot):
        self._snapshot = snapshot

        self._snapshot.save()
        self._plot_background()
        self._plot_header()
        self._plot_primary_fields()
        self._plot_secondary_and_axiliary_fields()
        self._snapshot.restore()

    def _plot_header(self):
        header_height = 32

        # Draw the logo if it exists
        if self._logo_texture:
            logo_scale = header_height / self._logo_texture.get_height()
            logo_width = self._logo_texture.get_width() * logo_scale

            rectangle = Graphene.Rect()
            rectangle.init(PASS_MARGIN, PASS_MARGIN, logo_width, header_height)
            self._snapshot.append_texture(self._logo_texture, rectangle)

        point = Graphene.Point()
        point.y = PASS_MARGIN

        right_margin = (PASS_WIDTH - PASS_MARGIN)

        self._snapshot.save()
        self._snapshot.translate(point)

        for field in self._header_fields:
            field_layout = FieldLayout(self._pango_context, field,
                                       alignment = Pango.Alignment.RIGHT)

            field_original_width = field_layout.get_width()
            field_layout.set_width(right_margin)

            field_layout.append(self._snapshot,
                                self._label_color,
                                self._fg_color)

            right_margin -= field_original_width + PASS_MARGIN

        self._snapshot.restore()

        # Perform a translation so that the next drawing starts below this one
        point.x = 0
        point.y = header_height + 3 * PASS_MARGIN
        self._snapshot.translate(point)

    def _plot_primary_fields(self):
        raise NotImplementedError

    def _plot_secondary_and_axiliary_fields(self):
        raise NotImplementedError

    def _plot_footer(self):
        raise NotImplementedError


class PkPassWithStripPlotter(PkPassPlotter):
    """
    PkPassWithStripPlotter is a PkPassPlotter for PKPasses that may contain a
    strip image.
    """

    STRIP_IMAGE_MAX_HEIGHT = 123

    def __init__(self, pkpass, pkpass_widget):
        super().__init__(pkpass, pkpass_widget)

    def _plot_primary_fields(self):

        # Draw the strip

        strip_height = 0
        draw_strip = self._strip_texture and not self._background_texture

        if draw_strip:
            strip_scale = PASS_WIDTH / self._strip_texture.get_width()
            strip_height = self._strip_texture.get_height() * strip_scale

            rectangle = Graphene.Rect()
            rectangle.init(0, -PASS_MARGIN, PASS_WIDTH, strip_height)

            strip_height = min(self.STRIP_IMAGE_MAX_HEIGHT, strip_height)
            strip_area = Graphene.Rect()
            strip_area.init(0, -PASS_MARGIN, PASS_WIDTH, strip_height)

            self._snapshot.push_clip(strip_area)
            self._snapshot.append_texture(self._strip_texture, rectangle)
            self._snapshot.pop()

        # Draw the primary fields

        point = Graphene.Point()
        field_layout_height = 0

        if self._primary_fields:
            field_layout = FieldLayout(self._pango_context,
                                       self._primary_fields[0],
                                       value_font = self.PRIMARY_FIELD_VALUE_FONT)

            field_layout_height = field_layout.get_height()
            self._snapshot.save()

            point.x = PASS_MARGIN
            point.y = 0
            self._snapshot.translate(point)

            label_color = self._fg_color if draw_strip else self._label_color
            field_layout.append(self._snapshot, label_color, self._fg_color)

            self._snapshot.restore()

        # Perform a translation so that the next drawing starts below this one
        point.x = 0
        point.y = strip_height if strip_height > field_layout_height \
                               else field_layout_height + 2 * PASS_MARGIN
        self._snapshot.translate(point)


class BoardingPassPlotter(PkPassPlotter):

    def __init__(self, pkpass, pkpass_widget):
        super().__init__(pkpass, pkpass_widget)

    def _plot_primary_fields(self):

        # Origin
        origin_field = FieldLayout(self._pango_context,
                                   self._primary_fields[0],
                                   value_font = PassFont.biggest_value,
                                   alignment = Pango.Alignment.LEFT)

        # Destination
        destination_field = FieldLayout(self._pango_context,
                                        self._primary_fields[1],
                                        value_font = PassFont.biggest_value,
                                        alignment = Pango.Alignment.RIGHT)

        destination_field.set_width(PASS_WIDTH - 2 * PASS_MARGIN)
        self._snapshot.save()

        point = Graphene.Point()
        point.x = PASS_MARGIN
        point.y = 0
        self._snapshot.translate(point)

        origin_field.append(self._snapshot, self._label_color, self._fg_color)
        destination_field.append(self._snapshot, self._label_color, self._fg_color)

        self._snapshot.restore()

        # Perform a translation so that the next drawing starts below this one
        point.x = 0
        point.y = max(origin_field.get_height(), destination_field.get_height()) + 2 * PASS_MARGIN
        self._snapshot.translate(point)

    def _plot_secondary_and_axiliary_fields(self):
        self._plot_fields_layouts(self._auxiliary_fields)
        self._plot_fields_layouts(self._secondary_fields)


class CouponPlotter(PkPassWithStripPlotter):

    STRIP_IMAGE_MAX_HEIGHT = 144

    def __init__(self, pkpass, pkpass_widget):
        super().__init__(pkpass, pkpass_widget)

    def _plot_secondary_and_axiliary_fields(self):
        self._plot_fields_layouts(self._secondary_fields + \
                                  self._auxiliary_fields)


class EventTicketPlotter(PkPassWithStripPlotter):

    PRIMARY_FIELD_VALUE_FONT = PassFont.big_value
    STRIP_IMAGE_MAX_HEIGHT = 98

    def __init__(self, pkpass, pkpass_widget):
        super().__init__(pkpass, pkpass_widget)

    def _plot_background(self):

        if not self._strip_texture and self._background_texture:
            rectangle = Graphene.Rect()
            rectangle.init(-BACKGROUND_BLUR_RADIUS,
                           -BACKGROUND_BLUR_RADIUS,
                           PASS_WIDTH + 2 * BACKGROUND_BLUR_RADIUS,
                           PASS_HEIGHT + 2 * BACKGROUND_BLUR_RADIUS)

            self._snapshot.push_blur(BACKGROUND_BLUR_RADIUS)
            self._snapshot.append_texture(self._background_texture, rectangle)
            self._snapshot.pop()

        else:
            super()._plot_background()

    def _plot_secondary_and_axiliary_fields(self):
        self._plot_fields_layouts(self._secondary_fields + \
                                  self._auxiliary_fields)



class GenericPlotter(PkPassPlotter):

    def __init__(self, pkpass, pkpass_widget):
        super().__init__(pkpass, pkpass_widget)

    def _plot_primary_fields(self):
        if not self._primary_fields:
            return

        field_layout = FieldLayout(self._pango_context,
                                   self._primary_fields[0],
                                   value_font = PassFont.big_value)
        self._snapshot.save()

        point = Graphene.Point()
        point.x = PASS_MARGIN
        point.y = 0
        self._snapshot.translate(point)

        field_layout.append(self._snapshot, self._label_color, self._fg_color)

        self._snapshot.restore()

        # Perform a translation so that the next drawing starts below this one
        point.x = 0
        point.y = field_layout.get_height() + 2 * PASS_MARGIN
        self._snapshot.translate(point)

    def _plot_secondary_and_axiliary_fields(self):
        self._plot_fields_layouts(self._auxiliary_fields)
        self._plot_fields_layouts(self._secondary_fields)


class PassWidget(Gtk.Fixed):

    __gtype_name__ = 'PassWidget'

    def __init__(self):
        super().__init__()

        self.__pass_plotter = None
        self.__barcode_button = None
        self.__children = []

        self.props.width_request = PASS_WIDTH
        self.props.height_request = PASS_HEIGHT

        self.props.hexpand = False
        self.props.vexpand = True
        self.props.halign = Gtk.Align.CENTER
        self.props.valign = Gtk.Align.CENTER

        self.props.focusable = True

        self.add_css_class('card')

    def __on_barcode_clicked(self, args):
        self.emit('barcode_clicked')

    @GObject.Signal
    def barcode_clicked(self):
        pass

    def do_snapshot(self, snapshot):
        if not self.__pass_plotter:
            return

        self.__pass_plotter.plot(snapshot)

        if self.__barcode_button:
            self.snapshot_child(self.__barcode_button, snapshot)

    def content(self, a_pass):
        if self.__barcode_button:
            self.remove(self.__barcode_button)
            self.__barcode_button = None

        self.__pass_plotter = PassPlotter.new(a_pass, self)
        self.create_barcode_button(a_pass)

        # After changing the plotter, we have to redraw the widget
        self.queue_draw()

    def create_barcode_button(self, a_pass):
        barcode = a_pass.barcodes()[0]

        if not barcode:
            return

        self.__barcode_button = Gtk.Button()
        self.__barcode_button.connect('clicked', self.__on_barcode_clicked)
        self.__barcode_button.add_css_class('barcode-button')
        barcode_widget = BarcodeWidget()
        barcode_widget.encode(barcode.format(),
                              barcode.message(),
                              barcode.message_encoding())

        aspect_ratio = barcode_widget.aspect_ratio()

        # Square codes
        if aspect_ratio == 1:
            max_times = 140 // barcode_widget.minimum_height()

            barcode_button_width = max_times * barcode_widget.minimum_height()
            barcode_button_height = max_times * barcode_widget.minimum_height()

        # Horizontal codes
        elif aspect_ratio > 1:
            max_times = (PASS_WIDTH - 2*PASS_MARGIN) // barcode_widget.minimum_width()

            if max_times * barcode_widget.minimum_height() > 140:
                max_times = 140 // barcode_widget.minimum_height()

            barcode_button_width = max_times * barcode_widget.minimum_width()
            barcode_button_height = max_times * barcode_widget.minimum_height()

        # Vertical codes
        else:
            barcode_button_width = 177
            barcode_button_height = 177

        self.__barcode_button.props.width_request = barcode_button_width
        self.__barcode_button.props.height_request = barcode_button_height

        self.__barcode_button.set_child(barcode_widget)
        self.put(self.__barcode_button,
                 PASS_WIDTH/2 - barcode_button_width/2,
                 PASS_HEIGHT - PASS_MARGIN - barcode_button_height)



# pkpass_back_view.py
#
# Copyright 2022-2023 Pablo Sánchez Rodríguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Adw, Gtk

from .pass_field_row import PassFieldRow


@Gtk.Template(resource_path='/me/sanchezrodriguez/passes/additional_information_pane.ui')
class AdditionalInformationPane(Gtk.Box):

    __gtype_name__ = 'AdditionalInformationPane'

    scrolled_window = Gtk.Template.Child()
    fields = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        # Create a placeholder widget to be displayed when the pane is empty
        placeholder = Adw.StatusPage.new()
        placeholder.set_icon_name('info-symbolic')
        placeholder.set_title(_('No additional information'))
        placeholder.add_css_class('compact')
        self.fields.set_placeholder(placeholder)

    def clean(self):
        # Remove all fields
        row = self.fields.get_row_at_index(0)
        while row:
            self.fields.remove(row)
            row = self.fields.get_row_at_index(0)

        # Move the scrollbar to the top
        vertical_adjustment = Gtk.Adjustment()
        vertical_adjustment.set_value(vertical_adjustment.get_lower())
        self.scrolled_window.set_vadjustment(vertical_adjustment)

    def content(self, a_pass):
        self.clean()
        fields = a_pass.additional_information()

        if len(fields) == 0:
            alignment = Gtk.Align.FILL
            if self.fields.has_css_class('boxed-list'):
                self.fields.remove_css_class('boxed-list')
        else:
            alignment = Gtk.Align.START
            if not self.fields.has_css_class('boxed-list'):
                self.fields.add_css_class('boxed-list')

        self.fields.set_valign(alignment)

        for field in fields:
            label = field.label()
            value = field.value()

            passFieldRow = PassFieldRow()
            passFieldRow.set_label(label)
            passFieldRow.set_value(value)
            self.fields.append(passFieldRow)
